// Copyright (C) 2022-2025 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef ISC_LIMITS_LOGGER_H
#define ISC_LIMITS_LOGGER_H

#include <log/logger_support.h>
#include <log/macros.h>

#include <limits/limits_messages.h>

namespace isc {
namespace limits {

extern isc::log::Logger limits_logger;

}  // namespace limits
}  // namespace isc

#endif  // ISC_LIMITS_LOGGER_H
