// Copyright (C) 2015-2024 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef EVAL_CONTEXT_DECL_H
#define EVAL_CONTEXT_DECL_H

/// @file eval_context_decl.h Forward declaration of the EvalContext class

namespace isc {
namespace eval {

class EvalContext;

}  // end of isc::eval namespace
}  // end of isc namespace

#endif
