// File created from src/hooks/dhcp/stat_cmds/stat_cmds_messages.mes

#ifndef STAT_CMDS_MESSAGES_H
#define STAT_CMDS_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID STAT_CMDS_DEINIT_OK;
extern const isc::log::MessageID STAT_CMDS_INIT_OK;
extern const isc::log::MessageID STAT_CMDS_LEASE4_FAILED;
extern const isc::log::MessageID STAT_CMDS_LEASE4_GET;
extern const isc::log::MessageID STAT_CMDS_LEASE4_GET_FAILED;
extern const isc::log::MessageID STAT_CMDS_LEASE4_GET_INVALID;
extern const isc::log::MessageID STAT_CMDS_LEASE4_GET_NO_SUBNETS;
extern const isc::log::MessageID STAT_CMDS_LEASE4_ORPHANED_STATS;
extern const isc::log::MessageID STAT_CMDS_LEASE6_FAILED;
extern const isc::log::MessageID STAT_CMDS_LEASE6_GET;
extern const isc::log::MessageID STAT_CMDS_LEASE6_GET_FAILED;
extern const isc::log::MessageID STAT_CMDS_LEASE6_GET_INVALID;
extern const isc::log::MessageID STAT_CMDS_LEASE6_GET_NO_SUBNETS;
extern const isc::log::MessageID STAT_CMDS_LEASE6_ORPHANED_STATS;

#endif // STAT_CMDS_MESSAGES_H
