// File created from src/hooks/dhcp/user_chk/user_chk_messages.mes

#ifndef USER_CHK_MESSAGES_H
#define USER_CHK_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID USER_CHK_HOOK_LOAD_ERROR;
extern const isc::log::MessageID USER_CHK_HOOK_UNLOAD_ERROR;
extern const isc::log::MessageID USER_CHK_SUBNET4_SELECT_ERROR;
extern const isc::log::MessageID USER_CHK_SUBNET4_SELECT_REGISTRY_NULL;
extern const isc::log::MessageID USER_CHK_SUBNET6_SELECT_ERROR;
extern const isc::log::MessageID USER_CHK_SUBNET6_SELECT_REGISTRY_NULL;

#endif // USER_CHK_MESSAGES_H
