// File created from src/hooks/dhcp/host_cmds/host_cmds_messages.mes

#ifndef HOST_CMDS_MESSAGES_H
#define HOST_CMDS_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID HOST_CMDS_DEINIT_OK;
extern const isc::log::MessageID HOST_CMDS_INIT_FAILED;
extern const isc::log::MessageID HOST_CMDS_INIT_OK;
extern const isc::log::MessageID HOST_CMDS_RESERV_ADD;
extern const isc::log::MessageID HOST_CMDS_RESERV_ADD_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_ADD_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_DEL;
extern const isc::log::MessageID HOST_CMDS_RESERV_DEL_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_DEL_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_ALL;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_ALL_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_ALL_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ADDRESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ADDRESS_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ADDRESS_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_HOSTNAME;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_HOSTNAME_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_HOSTNAME_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ID;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ID_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_BY_ID_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_PAGE;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_PAGE_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_PAGE_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_GET_SUCCESS;
extern const isc::log::MessageID HOST_CMDS_RESERV_UPDATE;
extern const isc::log::MessageID HOST_CMDS_RESERV_UPDATE_FAILED;
extern const isc::log::MessageID HOST_CMDS_RESERV_UPDATE_SUCCESS;

#endif // HOST_CMDS_MESSAGES_H
