// File created from src/hooks/dhcp/run_script/run_script_messages.mes

#ifndef RUN_SCRIPT_MESSAGES_H
#define RUN_SCRIPT_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID RUN_SCRIPT_LOAD;
extern const isc::log::MessageID RUN_SCRIPT_LOAD_ERROR;
extern const isc::log::MessageID RUN_SCRIPT_UNLOAD;

#endif // RUN_SCRIPT_MESSAGES_H
