// File created from src/hooks/dhcp/flex_id/flex_id_messages.mes

#ifndef FLEX_ID_MESSAGES_H
#define FLEX_ID_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID FLEX_ID_EXPRESSION_EMPTY;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_EVALUATED;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_EVALUATED_NP;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_HEX;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_INVALID_JSON_TYPE;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_NOT_DEFINED;
extern const isc::log::MessageID FLEX_ID_EXPRESSION_PARSE_FAILED;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_APPLIED_ON_NA;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_APPLIED_ON_PD;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_ENABLED;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_JSON_TYPE;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_NOT_APPLIED_ON_NA;
extern const isc::log::MessageID FLEX_ID_IGNORE_IAID_NOT_APPLIED_ON_PD;
extern const isc::log::MessageID FLEX_ID_LOAD_ERROR;
extern const isc::log::MessageID FLEX_ID_REPLACE_CLIENT_ID_JSON_TYPE;
extern const isc::log::MessageID FLEX_ID_RESTORE_CLIENT_ID;
extern const isc::log::MessageID FLEX_ID_RESTORE_DUID;
extern const isc::log::MessageID FLEX_ID_UNLOAD;
extern const isc::log::MessageID FLEX_ID_USED_AS_CLIENT_ID;
extern const isc::log::MessageID FLEX_ID_USED_AS_DUID;

#endif // FLEX_ID_MESSAGES_H
