// File created from src/hooks/dhcp/subnet_cmds/subnet_cmds_messages.mes

#ifndef SUBNET_CMDS_MESSAGES_H
#define SUBNET_CMDS_MESSAGES_H

#include <log/message_types.h>

namespace isc {
namespace subnet_cmds {

extern const isc::log::MessageID SUBNET_CMDS_DEINIT_OK;
extern const isc::log::MessageID SUBNET_CMDS_INIT_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_INIT_OK;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_GET_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_LIST_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_SUBNET_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK4_SUBNET_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_GET_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_LIST_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_SUBNET_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK6_SUBNET_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_ADD;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_DEL;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_GET;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_GET_EMPTY;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_LIST;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_LIST_EMPTY;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_SUBNET_ADD;
extern const isc::log::MessageID SUBNET_CMDS_NETWORK_SUBNET_DEL;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_DELTA_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_DELTA_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_GET_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_LIST_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET4_UPDATE_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_DELTA_ADD_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_DELTA_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_DEL_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_GET_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_LIST_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET6_UPDATE_FAILED;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_ADD;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_DEL;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_GET;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_GET_EMPTY;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_LIST;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_LIST_EMPTY;
extern const isc::log::MessageID SUBNET_CMDS_SUBNET_UPDATE;

} // namespace subnet_cmds
} // namespace isc

#endif // SUBNET_CMDS_MESSAGES_H
