// File created from src/hooks/dhcp/class_cmds/class_cmds_messages.mes

#ifndef CLASS_CMDS_MESSAGES_H
#define CLASS_CMDS_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID CLASS_CMDS_CLASS_ADD;
extern const isc::log::MessageID CLASS_CMDS_CLASS_ADD_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_ADD_HANDLER_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_DEL;
extern const isc::log::MessageID CLASS_CMDS_CLASS_DEL_EMPTY;
extern const isc::log::MessageID CLASS_CMDS_CLASS_DEL_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_DEL_HANDLER_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_GET;
extern const isc::log::MessageID CLASS_CMDS_CLASS_GET_EMPTY;
extern const isc::log::MessageID CLASS_CMDS_CLASS_GET_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_GET_HANDLER_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_LIST;
extern const isc::log::MessageID CLASS_CMDS_CLASS_LIST_EMPTY;
extern const isc::log::MessageID CLASS_CMDS_CLASS_LIST_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_LIST_HANDLER_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_UPDATE;
extern const isc::log::MessageID CLASS_CMDS_CLASS_UPDATE_EMPTY;
extern const isc::log::MessageID CLASS_CMDS_CLASS_UPDATE_FAILED;
extern const isc::log::MessageID CLASS_CMDS_CLASS_UPDATE_HANDLER_FAILED;
extern const isc::log::MessageID CLASS_CMDS_DEINIT_OK;
extern const isc::log::MessageID CLASS_CMDS_INIT_FAILED;
extern const isc::log::MessageID CLASS_CMDS_INIT_OK;

#endif // CLASS_CMDS_MESSAGES_H
