// Copyright (C) 2017-2024 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef D2_PARSER_CONTEXT_DECL_H
#define D2_PARSER_CONTEXT_DECL_H

/// @file d2/parser_context_decl.h Forward declaration of the ParserContext class

namespace isc {
namespace d2 {

class D2ParserContext;

}  // end of isc::dhcp namespace
}  // end of isc namespace

#endif
