// File created from src/hooks/dhcp/ddns_tuning/ddns_tuning_messages.mes

#ifndef DDNS_TUNING_MESSAGES_H
#define DDNS_TUNING_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID DDNS_TUNING4_CALCULATED_HOSTNAME;
extern const isc::log::MessageID DDNS_TUNING4_PROCESS_ERROR;
extern const isc::log::MessageID DDNS_TUNING4_SKIPPING_DDNS;
extern const isc::log::MessageID DDNS_TUNING6_CALCULATED_HOSTNAME;
extern const isc::log::MessageID DDNS_TUNING6_PROCESS_ERROR;
extern const isc::log::MessageID DDNS_TUNING6_SKIPPING_DDNS;
extern const isc::log::MessageID DDNS_TUNING_GLOBAL_EXPR_SET;
extern const isc::log::MessageID DDNS_TUNING_LOAD_ERROR;
extern const isc::log::MessageID DDNS_TUNING_LOAD_OK;
extern const isc::log::MessageID DDNS_TUNING_SUBNET_EXPRESSION_PARSE;
extern const isc::log::MessageID DDNS_TUNING_SUBNET_EXPRESSION_PARSE_ERROR;
extern const isc::log::MessageID DDNS_TUNING_SUBNET_EXPR_CACHED;
extern const isc::log::MessageID DDNS_TUNING_UNLOAD;

#endif // DDNS_TUNING_MESSAGES_H
