// Copyright (C) 2019-2024 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#include <config.h>

#include <mysql_cb_log.h>

namespace isc {
namespace dhcp {

isc::log::Logger mysql_cb_logger("mysql-cb-hooks");

}  // namespace dhcp
}  // namespace isc
