// Copyright (C) 2024 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#include <config.h>
#include <hooks/hooks.h>

extern "C" {

/// @brief returns Kea hooks version.
int version() {
    return (KEA_HOOKS_VERSION);
}

}
