// File created from src/hooks/dhcp/forensic_log/legal_log_messages.mes

#ifndef LEGAL_LOG_MESSAGES_H
#define LEGAL_LOG_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID LEGAL_LOG_COMMAND_NO_LEGAL_STORE;
extern const isc::log::MessageID LEGAL_LOG_COMMAND_WRITE_ERROR;
extern const isc::log::MessageID LEGAL_LOG_DB_OPEN_CONNECTION_WITH_RETRY_FAILED;
extern const isc::log::MessageID LEGAL_LOG_LEASE4_NO_LEGAL_STORE;
extern const isc::log::MessageID LEGAL_LOG_LEASE4_WRITE_ERROR;
extern const isc::log::MessageID LEGAL_LOG_LEASE6_NO_LEGAL_STORE;
extern const isc::log::MessageID LEGAL_LOG_LEASE6_WRITE_ERROR;
extern const isc::log::MessageID LEGAL_LOG_LOAD_ERROR;
extern const isc::log::MessageID LEGAL_LOG_STORE_CLOSED;
extern const isc::log::MessageID LEGAL_LOG_STORE_CLOSE_ERROR;
extern const isc::log::MessageID LEGAL_LOG_STORE_OPEN;
extern const isc::log::MessageID LEGAL_LOG_STORE_OPENED;
extern const isc::log::MessageID LEGAL_LOG_SYSLOG;
extern const isc::log::MessageID LEGAL_LOG_SYSLOG_STORE_OPEN;
extern const isc::log::MessageID LEGAL_LOG_UNLOAD_ERROR;

#endif // LEGAL_LOG_MESSAGES_H
