// File created from src/lib/log/tests/log_test_messages.mes

#ifndef LOG_TEST_MESSAGES_H
#define LOG_TEST_MESSAGES_H

#include <log/message_types.h>

namespace isc {
namespace log {

extern const isc::log::MessageID LOG_LOCK_TEST_MESSAGE;

} // namespace log
} // namespace isc

#endif // LOG_TEST_MESSAGES_H
