// Copyright (C) 2021-2025 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#include <config.h>
#include <gss_tsig_log.h>

namespace isc {
namespace gss_tsig {

isc::log::Logger gss_tsig_logger("gss-tsig-hooks");

} // end of namespace isc::gss_tsig
} // end of namespace isc
