// File created from src/hooks/dhcp/bootp/bootp_messages.mes

#ifndef BOOTP_MESSAGES_H
#define BOOTP_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID BOOTP_BOOTP_QUERY;
extern const isc::log::MessageID BOOTP_LOAD;
extern const isc::log::MessageID BOOTP_PACKET_OPTIONS_SKIPPED;
extern const isc::log::MessageID BOOTP_PACKET_PACK;
extern const isc::log::MessageID BOOTP_PACKET_PACK_FAIL;
extern const isc::log::MessageID BOOTP_PACKET_UNPACK_FAILED;
extern const isc::log::MessageID BOOTP_UNLOAD;

#endif // BOOTP_MESSAGES_H
