// File created from src/hooks/dhcp/flex_option/flex_option_messages.mes

#ifndef FLEX_OPTION_MESSAGES_H
#define FLEX_OPTION_MESSAGES_H

#include <log/message_types.h>

extern const isc::log::MessageID FLEX_OPTION_LOAD_ERROR;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_ADD;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_CLIENT_CLASS;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_ERROR;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_REMOVE;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_SUB_ADD;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_SUB_CLIENT_CLASS;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_SUB_REMOVE;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_SUB_SUPERSEDE;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_SUPERSEDE;
extern const isc::log::MessageID FLEX_OPTION_PROCESS_VENDOR_ID_MISMATCH;
extern const isc::log::MessageID FLEX_OPTION_UNLOAD;

#endif // FLEX_OPTION_MESSAGES_H
